import 'package:flutter/material.dart';
import 'package:apply/core/plugin.dart';
import 'package:apply/core/plugins_service.dart';
import 'package:apply/elements/div.dart';
import 'package:apply/plugins/font-weight.dart';
import 'package:apply/plugins/font_family.dart';
import 'package:apply/plugins/font_size.dart';
import 'package:apply/plugins/height.dart';
import 'package:apply/plugins/letter_spacing.dart';
import 'package:apply/plugins/line_height.dart';
import 'package:apply/plugins/text_align.dart';
import 'package:apply/plugins/text_color.dart';
import 'package:apply/plugins/text_decoration.dart';
import 'package:apply/plugins/text_opacity.dart';
import 'package:apply/plugins/width.dart';

extension ExtGetPluginFromId<T extends Plugin> on Set<T> {
  T? getPluginById(String id) {
    var plugin = PluginsService.findById(id);
    // if(this == null){
    //   return null;
    // }
    if (this.contains(plugin)) {
      return this.where((e) => e.id == id).first;
    }

    return null;
  }

  T? getPluginByIdOrDefault(String id) => (this.getPluginById(id) ?? PluginsService.findById(id)) as T;

  bool removeById(String id) => this.remove(PluginsService.findById(id));

  List<Plugin> getInheritedPlugins() {
    var typesToInherit = [
      FontWeightP,
      FontFamily,
      FontSize,
      LineHeight,
      TextAlignP,
      TextColorP,
      TextOpacity,
      TextDecorationP,
      LetterSpacing,
    ];
    return this.where((element) => typesToInherit.contains(element.runtimeType)).toList(growable: false);
  }
}

extension ExtFlex on Flex {
  Flex copyWith({
    List<Widget> children = const [],
    MainAxisAlignment mainAxisAlignment = MainAxisAlignment.start,
    CrossAxisAlignment crossAxisAlignment = CrossAxisAlignment.center,
  }) =>
      direction == Axis.horizontal
          ? Row(
              children: children,
              mainAxisAlignment: mainAxisAlignment,
              mainAxisSize: this.mainAxisSize,
              crossAxisAlignment: crossAxisAlignment,
            )
          : Column(
              children: children,
              mainAxisAlignment: mainAxisAlignment,
              mainAxisSize: this.mainAxisSize,
              crossAxisAlignment: crossAxisAlignment,
            );
}

typedef Widget ComposableWidget(WidgetBuilder child);

class Pair<E, F> {
  final E? first;
  final F? last;

  Pair(this.first, this.last);

  @override
  String toString() => 'Pair{first: $first, last: $last}';
}

double radiansFromDegrees(double degrees) => degrees * 0.0174533;

extension HasDiv on Widget {
  bool hasDiv(BuildContext context) {
    if (this is Div) {
      return true;
    } else {

      if(this is StatelessWidget){
        var _widget = (this as StatelessWidget).build(context);
        return _widget is Div;
      }else{
        return false;
      }
      // return widget != null;
    }
  }

  Div? setOwnPlugins(BuildContext context) {
    assert(hasDiv(context));
    Div? widget;
    if (this is Div) {
      widget = this as Div;
    } else {
      var _widget = (this as StatelessWidget).build(context);
      if(_widget is Div){
        widget = _widget;
      }
      if (widget == null) {
        return null;
      }
    }

    widget.setAttributes(context);
    widget.ownPlugins = widget.attributes?.getAppliedPlugins();
    var options = widget.divOptions;
    double? width = options.width;
    double? height = options.height;
    var plugins = widget.ownPlugins;

    Width? widthPlugin = plugins?.getPluginById("width") as Width?;
    Height? heightPlugin = plugins?.getPluginById("height") as Height?;
    if (widthPlugin != null) {
      width = widthPlugin.getWidth();
    }

    if (heightPlugin != null) {
      height = heightPlugin.getHeight();
    }

    widget.widthAndHeightPair = Pair(width, height);
    return widget;
  }
}


extension GetTextStyle on BuildContext{
  Pair<TextAlign, TextStyle> getTextStyle() {
    Div? enclosingDiv = this.findAncestorWidgetOfExactType<Div>();
    if (enclosingDiv == null) {
      return throw "Invalid use of tekst";
    }
    return enclosingDiv.getTextStyle(this);
  }

}


extension GetDivTextStyle on Div{
  Pair<TextAlign, TextStyle> getTextStyle(BuildContext context) {
    Div enclosingDiv = this;

    TextDecoration? textDecoration;

    TextDecorationP? textDecorationP = enclosingDiv.plugins?.getPluginById("textDecoration") as TextDecorationP?;
    if (textDecorationP != null) {
      textDecoration = textDecorationP.getTextDecoration();
    }

    double? fontSize;
    double? lineHeight = 1.5;
    double? letterSpacing = 0;
    String? fontFamily;
    FontWeight? fontWeight;
    Color? textColor;
    TextAlign? textAlign;
    TextStyle textStyle = TextStyle(
      letterSpacing: letterSpacing,
      height: lineHeight,
      decoration: textDecoration,
    );

    FontSize fontSizePlugin = enclosingDiv.plugins?.getPluginByIdOrDefault("fontSize") as FontSize;
    FontFamily? fontFamilyPlugin = enclosingDiv.plugins?.getPluginById("fontFamily") as FontFamily?;
    FontWeightP? fontWeightPlugin = enclosingDiv.plugins?.getPluginById("fontWeight") as FontWeightP?;
    LineHeight? lineHeightPlugin = enclosingDiv.plugins?.getPluginById("lineHeight") as LineHeight?;
    LetterSpacing? letterSpacingPlugin = enclosingDiv.plugins?.getPluginById("letterSpacing") as LetterSpacing?;
    TextAlignP? textAlignPlugin = enclosingDiv.plugins?.getPluginById("textAlign") as TextAlignP?;
    TextColorP? textColorPlugin = enclosingDiv.plugins?.getPluginById("textColor") as TextColorP?;
    TextOpacity? textOpacityPlugin = enclosingDiv.plugins?.getPluginById("textOpacity") as TextOpacity?;

    // if (fontSizePlugin != null) {
    fontSize = fontSizePlugin.fontSize();
    textStyle = textStyle.copyWith(fontSize: fontSize);
    // }

    if (lineHeightPlugin != null) {
      lineHeight = lineHeightPlugin.getLineHeight(fontSize);
      textStyle = textStyle.copyWith(height: lineHeight);
    }

    if (letterSpacingPlugin != null) {
      letterSpacing = letterSpacingPlugin.getLetterSpacing(fontSize);
      textStyle = textStyle.copyWith(letterSpacing: letterSpacing);
    }

    if (fontFamilyPlugin != null) {
      fontFamily = fontFamilyPlugin.fontFamily();
      textStyle = textStyle.copyWith(fontFamily: fontFamily);
    }

    if (fontWeightPlugin != null) {
      fontWeight = fontWeightPlugin.getFontWeight();
      textStyle = textStyle.copyWith(fontWeight: fontWeight);
    }

    if (textColorPlugin != null) {
      textColor = textColorPlugin.getColor(enclosingDiv);
      textStyle = textStyle.copyWith(color: textColor);
    }

    if (textOpacityPlugin != null) {
      if(textColor == null){
        textColor = Theme.of(context).textTheme.bodyText1?.color;
      }
      textColor = textColor!.withOpacity(textOpacityPlugin.getColorOpacity());
      textStyle = textStyle.copyWith(color: textColor);
    }

    if (textAlignPlugin != null) {
      textAlign = textAlignPlugin.getTextAlign();
    }

    return Pair(textAlign, textStyle);
  }

}
